@@ -18,6 +18,7 @@
 
 #include "kerncompat.h"
 #include <sys/ioctl.h>
+#include <stdbool.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <pthread.h>
@@ -28,16 +29,20 @@
 #include <stdlib.h>
 #include <string.h>
 #include "kernel-lib/sizes.h"
+#include "kernel-shared/uapi/btrfs.h"
 #include "common/utils.h"
 #include "common/send-utils.h"
 #include "common/help.h"
 #include "common/path-utils.h"
+#include "common/sysfs-utils.h"
 #include "common/string-utils.h"
 #include "common/messages.h"
 #include "cmds/commands.h"
-#include "ioctl.h"
 
-#define SEND_BUFFER_SIZE	SZ_64K
+#define BTRFS_SEND_BUF_SIZE_V1	(SZ_64K)
+#define BTRFS_MAX_COMPRESSED	(SZ_128K)
+#define BTRFS_SEND_BUF_SIZE_V2	(SZ_16K + BTRFS_MAX_COMPRESSED)
+
 
 struct btrfs_send {
 	int send_fd;
@@ -201,11 +206,18 @@ static void *read_sent_data(void *arg)
 	struct btrfs_send *sctx = (struct btrfs_send*)arg;
 
 	while (1) {
+		size_t splice_buf_size = BTRFS_SEND_BUF_SIZE_V1;
 		ssize_t sbytes;
 
+		if(sctx->proto > 1){
+			splice_buf_size = BTRFS_SEND_BUF_SIZE_V2;
+
+			/* Try to change pipe buffer size */
+			fcntl(sctx->dump_fd, F_SETPIPE_SZ, splice_buf_size);
+		}
 		/* Source is a pipe, output is either file or stdout */
 		sbytes = splice(sctx->send_fd, NULL, sctx->dump_fd,
-				NULL, SEND_BUFFER_SIZE, SPLICE_F_MORE);
+				NULL, splice_buf_size, SPLICE_F_MORE);
 		if (sbytes < 0) {
 			ret = -errno;
 			error("failed to read stream from kernel: %m");
@@ -261,6 +273,9 @@ static int do_send(struct btrfs_send *send, u64 parent_root_id,
 		 */
 		io_send.version = send->proto;
 		io_send.flags |= BTRFS_SEND_FLAG_VERSION;
+
+		fcntl(pipefd[0], F_SETPIPE_SZ, BTRFS_SEND_BUF_SIZE_V2);
+		fcntl(pipefd[1], F_SETPIPE_SZ, BTRFS_SEND_BUF_SIZE_V2);
 	}
 
 	if (!ret)
@@ -303,8 +318,9 @@ static int do_send(struct btrfs_send *send, u64 parent_root_id,
 	}
 	if (t_err) {
 		ret = (long int)t_err;
-		error("failed to process send stream, ret=%ld (%s)",
-				(long int)t_err, strerror(-ret));
+		errno = -ret;
+		error("failed to process send stream, ret=%ld (%m)",
+				(long int)t_err);
 		goto out;
 	}
 
@@ -465,26 +481,20 @@ static const char * const cmd_send_usage[] = {
 	"which case 'btrfs send' will determine a suitable parent among the",
 	"clone sources itself.",
 	"",
-	"-e               If sending multiple subvols at once, use the new",
-	"                 format and omit the end-cmd between the subvols.",
-	"-p <parent>      Send an incremental stream from <parent> to",
-	"                 <subvol>.",
-	"-c <clone-src>   Use this snapshot as a clone source for an ",
-	"                 incremental send (multiple allowed)",
-	"-f <outfile>     Output is normally written to stdout. To write to",
-	"                 a file, use this option. An alternative would be to",
-	"                 use pipes.",
-	"--no-data        send in NO_FILE_DATA mode, Note: the output stream",
-	"                 does not contain any file data and thus cannot be used",
-	"                 to transfer changes. This mode is faster and useful to",
-	"                 show the differences in metadata.",
-	"--proto N        use protocol version N, or 0 to use the highest version",
-	"                 supported by the sending kernel (default: 1)",
-	"--compressed-data",
-	"                 send data that is compressed on the filesystem directly",
-	"                 without decompressing it",
-	"-v|--verbose     deprecated, alias for global -v option",
-	"-q|--quiet       deprecated, alias for global -q option",
+	OPTLINE("-e", "if sending multiple subvols at once, use the new format and omit the end-cmd between the subvols"),
+	OPTLINE("-p <parent>", "send an incremental stream from <parent> to <subvol>"),
+	OPTLINE("-c <clone-src>", "Use this snapshot as a clone source for an incremental send (multiple allowed)"),
+	OPTLINE("-f <outfile>", "Output is normally written to stdout. To write to "
+		"a file, use this option. An alternative would be to use pipes."),
+	OPTLINE("--no-data", "send in NO_FILE_DATA mode, Note: the output stream "
+		"does not contain any file data and thus cannot be used "
+		"to transfer changes. This mode is faster and useful to "
+		"show the differences in metadata."),
+	OPTLINE("--proto N", "use protocol version N, or 0 to use the highest version "
+		"supported by the sending kernel (default: 1)"),
+	OPTLINE("--compressed-data", "send data that is compressed on the filesystem directly without decompressing it"),
+	OPTLINE("-v|--verbose", "deprecated, alias for global -v option"),
+	OPTLINE("-q|--quiet", "deprecated, alias for global -q option"),
 	HELPINFO_INSERT_GLOBALS,
 	HELPINFO_INSERT_VERBOSE,
 	HELPINFO_INSERT_QUIET,
