@@ -183,27 +183,39 @@ int main(int argc, char **argv) {
 		u64 cycles;
 		u64 time;
 		unsigned long cpu_flag;
-		void (*init_accel)(void);
+		int backend;
 	} contestants[] = {
 		{ .name = "NULL-NOP", .digest = hash_null_nop, .digest_size = 32 },
 		{ .name = "NULL-MEMCPY", .digest = hash_null_memcpy, .digest_size = 32 },
 		{ .name = "CRC32C-ref", .digest = hash_crc32c, .digest_size = 4,
-		  .cpu_flag = CPU_FLAG_NONE, .init_accel = hash_init_crc32c },
+		  .cpu_flag = CPU_FLAG_NONE },
 		{ .name = "CRC32C-NI", .digest = hash_crc32c, .digest_size = 4,
-		  .cpu_flag = CPU_FLAG_SSE42, .init_accel = hash_init_crc32c },
+		  .cpu_flag = CPU_FLAG_PCLMUL },
 		{ .name = "XXHASH", .digest = hash_xxhash, .digest_size = 8 },
 		{ .name = "SHA256-ref", .digest = hash_sha256, .digest_size = 32,
-		  .cpu_flag = CPU_FLAG_NONE, .init_accel = hash_init_sha256 },
+		  .cpu_flag = CPU_FLAG_NONE, .backend = CRYPTOPROVIDER_BUILTIN + 1 },
+		{ .name = "SHA256-gcrypt", .digest = hash_sha256, .digest_size = 32,
+		  .cpu_flag = CPU_FLAG_NONE, .backend = CRYPTOPROVIDER_LIBGCRYPT + 1 },
+		{ .name = "SHA256-sodium", .digest = hash_sha256, .digest_size = 32,
+		  .cpu_flag = CPU_FLAG_NONE, .backend = CRYPTOPROVIDER_LIBSODIUM + 1 },
+		{ .name = "SHA256-kcapi", .digest = hash_sha256, .digest_size = 32,
+		  .cpu_flag = CPU_FLAG_NONE, .backend = CRYPTOPROVIDER_LIBKCAPI + 1 },
 		{ .name = "SHA256-NI", .digest = hash_sha256, .digest_size = 32,
-		  .cpu_flag = CPU_FLAG_SHA, .init_accel = hash_init_sha256 },
+		  .cpu_flag = CPU_FLAG_SHA, .backend = CRYPTOPROVIDER_BUILTIN + 1 },
 		{ .name = "BLAKE2-ref", .digest = hash_blake2b, .digest_size = 32,
-		  .cpu_flag = CPU_FLAG_NONE, .init_accel = hash_init_blake2 },
+		  .cpu_flag = CPU_FLAG_NONE, .backend = CRYPTOPROVIDER_BUILTIN + 1 },
+		{ .name = "BLAKE2-gcrypt", .digest = hash_blake2b, .digest_size = 32,
+		  .cpu_flag = CPU_FLAG_NONE, .backend = CRYPTOPROVIDER_LIBGCRYPT + 1 },
+		{ .name = "BLAKE2-sodium", .digest = hash_blake2b, .digest_size = 32,
+		  .cpu_flag = CPU_FLAG_NONE, .backend = CRYPTOPROVIDER_LIBSODIUM + 1 },
+		{ .name = "BLAKE2-kcapi", .digest = hash_blake2b, .digest_size = 32,
+		  .cpu_flag = CPU_FLAG_NONE, .backend = CRYPTOPROVIDER_LIBKCAPI + 1 },
 		{ .name = "BLAKE2-SSE2", .digest = hash_blake2b, .digest_size = 32,
-		  .cpu_flag = CPU_FLAG_SSE2, .init_accel = hash_init_blake2 },
+		  .cpu_flag = CPU_FLAG_SSE2, .backend = CRYPTOPROVIDER_BUILTIN + 1 },
 		{ .name = "BLAKE2-SSE41", .digest = hash_blake2b, .digest_size = 32,
-		  .cpu_flag = CPU_FLAG_SSE41, .init_accel = hash_init_blake2 },
+		  .cpu_flag = CPU_FLAG_SSE41, .backend = CRYPTOPROVIDER_BUILTIN + 1 },
 		{ .name = "BLAKE2-AVX2", .digest = hash_blake2b, .digest_size = 32,
-		  .cpu_flag = CPU_FLAG_AVX2, .init_accel = hash_init_blake2 },
+		  .cpu_flag = CPU_FLAG_AVX2, .backend = CRYPTOPROVIDER_BUILTIN + 1 },
 	};
 	int units = UNITS_CYCLES;
 
@@ -270,15 +282,18 @@ int main(int argc, char **argv) {
 		u64 total = 0;
 
 		if (c->cpu_flag != 0 && !cpu_has_feature(c->cpu_flag)) {
-			printf("%12s: no CPU support\n", c->name);
+			printf("%14s: no CPU support\n", c->name);
 			continue;
 		}
-		printf("%12s: ", c->name);
+		/* Backend not compiled in */
+		if (c->backend == 1)
+			continue;
+		printf("%14s: ", c->name);
 		fflush(stdout);
 
 		if (c->cpu_flag) {
 			cpu_set_level(c->cpu_flag);
-			c->init_accel();
+			hash_init_accel();
 		}
 		tstart = get_time();
 		start = get_cycles(units);
